@@ -1,26 +1,44 @@
 """
-This module provides the AST. Subclass Context and override the various
-methods to allow minivect visitors over the AST, to promote and map types,
-etc. Subclass and override ASTBuilder's method to provide alternative
+This module provides the AST. Subclass :py:class:`Context` and override the
+various methods to allow minivect visitors over the AST, to promote and map types,
+etc. Subclass and override :py:class:`ASTBuilder`'s methods to provide alternative
 AST nodes or different implementations.
 """
 
 import copy
+import string
 import types
 
 import minitypes
 import miniutils
 import minivisitor
+import specializers
+import type_promoter
 import minicode
 import codegen
+import llvm_codegen
+import graphviz
+
+try:
+    import llvm.core
+    import llvm.ee
+    import llvm.passes
+except ImportError:
+    llvm = None
 
 class UndocClassAttribute(object):
+    "Use this to document class attributes for Sphinx"
     def __init__(self, cls):
         self.cls = cls
 
     def __call__(self, *args, **kwargs):
         return self.cls(*args, **kwargs)
 
+def make_cls(cls1, cls2):
+    "Fuse two classes together."
+    name = "%s_%s" % (cls1.__name__, cls2.__name__)
+    return type(name, (cls1, cls2), {})
+
 class Context(object):
     """
     A context that knows how to map ASTs back and forth, how to wrap nodes
@@ -69,26 +87,74 @@ class Context(object):
         functionality. This class should likely participate
         cooperatively in MI.
 
+    .. attribute:: variable_resolving_mixin_cls
+
+        A specializer mixin class that resolves wrapped miniasts in a foreign
+        AST. This is only needed if you are using :py:class:`NodeWrapper`,
+        which wraps a miniast somewhere at the leaves.
+
+    .. attribute: graphviz_cls
+
+        Visitor to generate a Graphviz graph. See the :py:module:`graphviz`
+        module.
+
+    .. attribute: minifunction
+
+        The current minifunction that is being translated.
+
     Use subclass :py:class:`CContext` to get the defaults for C code generation.
     """
 
     debug = False
+    debug_elements = False
 
-    codegen_cls = UndocClassAttribute(codegen.CodeGen)
+    use_llvm = False
+    optimize_broadcasting = True
+
+    shape_type = minitypes.Py_ssize_t.pointer()
+    strides_type = shape_type
+
+    astbuilder_cls = None
+    codegen_cls = UndocClassAttribute(codegen.VectorCodegen)
     cleanup_codegen_cls = UndocClassAttribute(codegen.CodeGenCleanup)
     codewriter_cls = UndocClassAttribute(minicode.CodeWriter)
     codeformatter_cls = UndocClassAttribute(minicode.CodeFormatter)
+    graphviz_cls = UndocClassAttribute(graphviz.GraphvizGenerator)
 
     specializer_mixin_cls = None
+    variable_resolving_mixin_cls = None
 
-    def __init__(self, astbuilder=None, typemapper=None):
-        self.astbuilder = astbuilder or ASTBuilder(self)
-        self.typemapper = typemapper or minitypes.TypeMapper(self)
+    func_counter = 0
+
+    final_specializer = specializers.FinalSpecializer
+
+    def __init__(self):
+        self.init()
+        if self.use_llvm:
+            if llvm is None:
+                import llvm.core as llvm_py_not_available # llvm-py not available
+
+            self.llvm_module = llvm.core.Module.new('default_module')
+            # self.llvm_ee = llvm.ee.ExecutionEngine.new(self.llvm_module)
+            self.llvm_ee = llvm.ee.EngineBuilder.new(self.llvm_module).force_jit().opt(3).create()
+            self.llvm_fpm = llvm.passes.FunctionPassManager.new(self.llvm_module)
+            self.llvm_fpm.initialize()
+            if not self.debug:
+                for llvm_pass in self.llvm_passes():
+                    self.llvm_fpm.add(llvm_pass)
+        else:
+            self.llvm_ee = None
+            self.llvm_module = None
+
+    def init(self):
+        self.astbuilder = self.astbuilder_cls(self)
+        self.typemapper = minitypes.TypeMapper(self)
 
     def run_opaque(self, astmapper, opaque_ast, specializers):
         return self.run(astmapper.visit(opaque_ast), specializers)
 
-    def run(self, ast, specializer_classes):
+    def run(self, ast, specializer_classes, graphviz_outfile=None,
+            print_tree=False):
         """
         Specialize the given AST with all given specializers and return
         an iterable of generated code in the form of
@@ -98,20 +164,68 @@ class Context(object):
         depending on the code formatter used.
         """
         for specializer_class in specializer_classes:
-            if self.specializer_mixin_cls:
-                cls1, cls2 = self.specializer_mixin_cls, specializer_class
-                name = "%s_%s" % (cls1.__name__, cls2.__name__)
-                specializer_class = type(name, (cls1, cls2), {})
+            self.init()
+            pipeline = self.pipeline(specializer_class)
+
+            specialized_ast = specializers.specialize_ast(ast)
+            self.astbuilder.minifunction = specialized_ast
+            for transform in pipeline:
+                specialized_ast = transform.visit(specialized_ast)
+
+            if print_tree:
+                specialized_ast.print_tree(self)
+
+            if graphviz_outfile is not None:
+                data = self.graphviz(specialized_ast)
+                graphviz_outfile.write(data)
 
-            specializer = specializer_class(self)
-            specialized_ast = specializer.visit(copy.deepcopy(ast))
-            # specialized_ast.print_tree(self)
             codewriter = self.codewriter_cls(self)
-            visitor = self.codegen_cls(self, codewriter)
-            visitor.visit(specialized_ast)
-            yield (specializer, specialized_ast, codewriter,
+            codegen = self.codegen_cls(self, codewriter)
+            codegen.visit(specialized_ast)
+
+            yield (pipeline[0], specialized_ast, codewriter,
                    self.codeformatter_cls().format(codewriter))
 
+    def debug_c(self, ast, specializer, astbuilder_cls=None):
+        "Generate C code (for debugging)"
+        context = CContext()
+
+        if astbuilder_cls:
+            context.astbuilder_cls = astbuilder_cls
+        else:
+            context.astbuilder_cls = self.astbuilder_cls
+
+        context.shape_type = self.shape_type
+        context.strides_type = self.strides_type
+
+        context.debug = self.debug
+        result = context.run(ast, [specializer]).next()
+        _, specialized_ast, _, (proto, impl) = result
+        return impl
+
+    def pipeline(self, specializer_class):
+        # add specializer mixin and run specializer
+        if self.specializer_mixin_cls:
+            specializer_class = make_cls(self.specializer_mixin_cls,
+                                         specializer_class)
+
+        specializer = specializer_class(self)
+        pipeline = [specializer]
+
+        # Add variable resolving mixin to the final specializer and run
+        # transform
+        final_specializer_cls = self.final_specializer
+        if final_specializer_cls:
+            if self.variable_resolving_mixin_cls:
+                final_specializer_cls = make_cls(
+                    self.variable_resolving_mixin_cls,
+                    final_specializer_cls)
+
+            pipeline.append(final_specializer_cls(self, specializer))
+
+        pipeline.append(type_promoter.TypePromoter(self))
+        return pipeline
+
     def generate_disposal_code(self, code, node):
         "Run the disposal code generator on an (sub)AST"
         transform = self.cleanup_codegen_cls(self, code)
@@ -121,6 +235,39 @@ class Context(object):
     ### Override in subclasses where needed
     #
 
+    def llvm_passes(self):
+        "Returns a list of LLVM optimization passes"
+        return []
+        return [
+            # llvm.passes.PASS_CFG_SIMPLIFICATION
+            llvm.passes.PASS_BLOCK_PLACEMENT,
+            llvm.passes.PASS_BASIC_ALIAS_ANALYSIS,
+            llvm.passes.PASS_NO_AA,
+            llvm.passes.PASS_SCALAR_EVOLUTION_ALIAS_ANALYSIS,
+#            llvm.passes.PASS_ALIAS_ANALYSIS_COUNTER,
+            llvm.passes.PASS_AAEVAL,
+            llvm.passes.PASS_LOOP_DEPENDENCE_ANALYSIS,
+            llvm.passes.PASS_BREAK_CRITICAL_EDGES,
+            llvm.passes.PASS_LOOP_SIMPLIFY,
+            llvm.passes.PASS_PROMOTE_MEMORY_TO_REGISTER,
+            llvm.passes.PASS_CONSTANT_PROPAGATION,
+            llvm.passes.PASS_LICM,
+            # llvm.passes.PASS_CONSTANT_MERGE,
+            llvm.passes.PASS_LOOP_STRENGTH_REDUCE,
+            llvm.passes.PASS_LOOP_UNROLL,
+            # llvm.passes.PASS_FUNCTION_ATTRS,
+            # llvm.passes.PASS_GLOBAL_OPTIMIZER,
+            # llvm.passes.PASS_GLOBAL_DCE,
+            llvm.passes.PASS_DEAD_CODE_ELIMINATION,
+            llvm.passes.PASS_INSTRUCTION_COMBINING,
+            llvm.passes.PASS_CODE_GEN_PREPARE,
+        ]
+
+    def mangle_function_name(self, name):
+        name = "%s_%d" % (name, self.func_counter)
+        self.func_counter += 1
+        return name
+
     def promote_types(self, type1, type2):
         "Promote types in an arithmetic operation"
         return self.typemapper.promote_types(type1, type2)
@@ -150,14 +297,24 @@ class Context(object):
         "Return an LLVM type for the given minitype"
         return self.typemapper.to_llvm(type)
 
+    def graphviz(self, node, graphviz_name="AST"):
+        visitor = self.graphviz_cls(self, graphviz_name)
+        graphviz_graph = visitor.visit(node)
+        return graphviz_graph.to_string()
 
 class CContext(Context):
     "Set defaults for C code generation."
 
-    codegen_cls = codegen.CCodeGen
+    codegen_cls = codegen.VectorCodegen
     codewriter_cls = minicode.CCodeWriter
     codeformatter_cls = minicode.CCodeStringFormatter
 
+class LLVMContext(Context):
+    "Context with default for LLVM code generation"
+
+    use_llvm = True
+    codegen_cls = llvm_codegen.LLVMCodeGen
+
 class ASTBuilder(object):
     """
     This class is used to build up a minivect AST. It can be used by a user
@@ -169,6 +326,7 @@ class ASTBuilder(object):
     # the 'pos' attribute is set for each visit to each node by
     # the ASTMapper
     pos = None
+    temp_reprname_counter = 0
 
     def __init__(self, context):
         """
@@ -187,7 +345,22 @@ class ASTBuilder(object):
         else:
             raise minierror.InferTypeError()
 
-    def function(self, name, body, args, shapevar=None, posinfo=None):
+    def create_function_type(self, function, strides_args=True):
+        arg_types = []
+        for arg in function.arguments + function.scalar_arguments:
+            if arg.used:
+                if arg.type and arg.type.is_array and not strides_args:
+                    arg_types.append(arg.data_pointer.type)
+                    arg.variables = [arg.data_pointer]
+                else:
+                    for variable in arg.variables:
+                        arg_types.append(variable.type)
+
+        function.type = minitypes.FunctionType(
+                    return_type=function.success_value.type, args=arg_types)
+
+    def function(self, name, body, args, shapevar=None, posinfo=None,
+                 omp_size=None):
         """
         Create a new function.
 
@@ -206,8 +379,7 @@ class ASTBuilder(object):
                         arguments to the function ``(filename, lineno, column)``.
         """
         if shapevar is None:
-            shapevar = self.variable(minitypes.Py_ssize_t.pointer(),
-                                     '__pyx_shape')
+            shapevar = self.variable(self.context.shape_type, 'shape')
 
         arguments, scalar_arguments = [], []
         for arg in args:
@@ -219,17 +391,47 @@ class ASTBuilder(object):
         arguments.insert(0, self.funcarg(shapevar))
         if posinfo:
             arguments.insert(1, posinfo)
-        body = self.nditerate(body)
-        return FunctionNode(self.pos, name, body, arguments, scalar_arguments,
-                            shapevar, posinfo,
-                            error_value=self.constant(-1),
-                            success_value=self.constant(0))
 
-    def funcarg(self, variable, *variables, **kwargs):
+        body = self.stats(self.nditerate(body))
+
+        error_value = self.constant(-1)
+        success_value = self.constant(0)
+
+        function = FunctionNode(self.pos, name, body,
+                            arguments, scalar_arguments,
+                            shapevar, posinfo,
+                            error_value=error_value,
+                            success_value=success_value,
+                            omp_size=omp_size or self.constant(1024))
+
+        # prepending statements, used during specialization
+        function.prepending = self.stats()
+        function.body = self.stats(function.prepending, function.body)
+
+        self.create_function_type(function)
+        return function
+
+    def build_function(self, variables, body, name=None, shapevar=None):
+        "Convenience method for building a minivect function"
+        args = []
+        for var in variables:
+            if var.type.is_array:
+                args.append(self.array_funcarg(var))
+            else:
+                args.append(self.funcarg(var))
+
+        name = name or 'function'
+        return self.function(name, body, args, shapevar=shapevar)
+
+    def funcarg(self, variable, *variables):
         """
         Create a (compound) function argument consisting of one or multiple
         argument Variables.
         """
+        if variable.type is not None and variable.type.is_array:
+            assert not variables
+            return self.array_funcarg(variable)
+
         if not variables:
             variables = [variable]
         return FunctionArgument(self.pos, variable, list(variables))
@@ -258,7 +460,7 @@ class ASTBuilder(object):
         "Print out all arguments to stdout"
         return PrintNode(self.pos, args=list(args))
 
-    def funccall(self, func_or_pointer, args):
+    def funccall(self, func_or_pointer, args, inline=False):
         """
         Generate a call to the given function (a :py:class:`FuncNameNode`) of
         :py:class:`minivect.minitypes.FunctionType` or a
@@ -267,11 +469,13 @@ class ASTBuilder(object):
         type = func_or_pointer.type
         if type.is_pointer:
             type = func_or_pointer.type.base_type
-        return FuncCallNode(self.pos, type,
-                            func_or_pointer=func_or_pointer, args=args)
+        return FuncCallNode(self.pos, type.return_type,
+                            func_or_pointer=func_or_pointer, args=args,
+                            inline=inline)
 
-    def funcname(self, type, name):
-        return FuncNameNode(self.pos, type, name=name)
+    def funcname(self, type, name, is_external=True):
+        assert type.is_function
+        return FuncNameNode(self.pos, type, name=name, is_external=is_external)
 
     def nditerate(self, body):
         """
@@ -281,7 +485,7 @@ class ASTBuilder(object):
         """
         return NDIterate(self.pos, body)
 
-    def for_(self, body, init, condition, step, is_tiled=False):
+    def for_(self, body, init, condition, step, index=None):
         """
         Create a for loop node.
 
@@ -290,7 +494,7 @@ class ASTBuilder(object):
         :param condition: boolean loop condition
         :param step: step clause (assignment expression)
         """
-        return ForNode(self.pos, init, condition, step, body, is_tiled)
+        return ForNode(self.pos, init, condition, step, body, index=index)
 
     def for_range_upwards(self, body, upper, lower=None, step=None):
         """
@@ -300,12 +504,14 @@ class ASTBuilder(object):
         :param body: the loop body
         :param upper: expression specifying an upper bound
         """
-        if lower is None:
-            lower = self.constant(0)
-        if step is None:
-            step = self.constant(1)
+        index_type = upper.type.unqualify("const")
 
-        temp = self.temp(minitypes.Py_ssize_t)
+        if lower is None:
+            lower = self.constant(0, index_type)
+        if step is None:
+            step = self.constant(1, index_type)
+
+        temp = self.temp(index_type)
         init = self.assign_expr(temp, lower)
         condition = self.binop(minitypes.bool_, '<', temp, upper)
         step = self.assign_expr(temp, self.add(temp, step))
@@ -323,8 +529,16 @@ class ASTBuilder(object):
                           executed sequentially (to avoid synchronization
                           overhead)
         """
+        if isinstance(for_node, PragmaForLoopNode):
+            for_node = for_node.for_node
         return OpenMPLoopNode(self.pos, for_node=for_node,
-                              if_clause=if_clause)
+                              if_clause=if_clause,
+                              lastprivates=[for_node.init.lhs],
+                              privates=[])
+
+    def omp_if(self, if_body, else_body=None):
+        return OpenMPConditionalNode(self.pos, if_body=if_body,
+                                     else_body=else_body)
 
     def pragma_for(self, for_node):
         """
@@ -336,14 +550,7 @@ class ASTBuilder(object):
         """
         Wrap a bunch of statements in an AST node.
         """
-        stats = []
-        for stat in statements:
-            if stat.is_statlist:
-                stats.extend(stat.stats)
-            else:
-                stats.append(stat)
-
-        return StatListNode(self.pos, stats)
+        return StatListNode(self.pos, list(statements))
 
     def expr_stat(self, expr):
         "Turn an expression into a statement"
@@ -356,13 +563,25 @@ class ASTBuilder(object):
 
     def if_(self, cond, body):
         "If statement"
-        return IfNode(self.pos, cond=cond, body=body)
+        return self.if_else(cond, body, None)
 
     def if_else_expr(self, cond, lhs, rhs):
         "If/else expression, resulting in lhs if cond else rhs"
         type = self.context.promote_types(lhs.type, rhs.type)
         return IfElseExprNode(self.pos, type=type, cond=cond, lhs=lhs, rhs=rhs)
 
+    def if_else(self, cond, if_body, else_body):
+        return IfNode(self.pos, cond=cond, body=if_body, else_body=else_body)
+
+    def promote(self, dst_type, node):
+        "Promote or demote the node to the given dst_type"
+        if node.type != dst_type:
+            if node.is_constant and node.type.kind == dst_type.kind:
+                node.type = dst_type
+                return node
+            return PromotionNode(self.pos, dst_type, node)
+        return node
+
     def binop(self, type, op, lhs, rhs):
         """
         Binary operation on two nodes.
@@ -373,7 +592,7 @@ class ASTBuilder(object):
         """
         return BinopNode(self.pos, type, op, lhs, rhs)
 
-    def add(self, lhs, rhs, result_type=None):
+    def add(self, lhs, rhs, result_type=None, op='+'):
         """
         Shorthand for the + binop. Filters out adding 0 constants.
         """
@@ -384,7 +603,10 @@ class ASTBuilder(object):
 
         if result_type is None:
             result_type = self.context.promote_types(lhs.type, rhs.type)
-        return self.binop(result_type, '+', lhs, rhs)
+        return self.binop(result_type, op, lhs, rhs)
+
+    def sub(self, lhs, rhs, result_type=None):
+        return self.add(lhs, rhs, result_type, op='-')
 
     def mul(self, lhs, rhs, result_type=None, op='*'):
         """
@@ -439,16 +661,18 @@ class ASTBuilder(object):
 
         return self.dereference(pointer)
 
-    def assign_expr(self, node, value):
+    def assign_expr(self, node, value, may_reorder=False):
         "Create an assignment expression assigning ``value`` to ``node``"
         assert node is not None
         if not isinstance(value, Node):
             value = self.constant(value)
-        return AssignmentExpr(self.pos, node.type, node, value)
+        return AssignmentExpr(self.pos, node.type, node, value,
+                              may_reorder=may_reorder)
 
-    def assign(self, node, value):
+    def assign(self, node, value, may_reorder=False):
         "Assignment statement"
-        return self.expr_stat(self.assign_expr(node, value))
+        expr = self.assign_expr(node, value, may_reorder=may_reorder)
+        return self.expr_stat(expr)
 
     def dereference(self, pointer):
         "Dereference a pointer"
@@ -466,9 +690,13 @@ class ASTBuilder(object):
         temp = self.temp(type)
         return self.expr(stats=[self.assign(temp, expr)], expr=temp)
 
-    def temp(self, type, name=None, rhs=None):
+    def temp(self, type, name=None):
         "Allocate a temporary of a given type"
-        return TempNode(self.pos, type, name=name or 'temp', rhs=rhs)
+        name = name or 'temp'
+        repr_name = '%s%d' % (name.rstrip(string.digits),
+                              self.temp_reprname_counter)
+        self.temp_reprname_counter += 1
+        return TempNode(self.pos, type, name=name, repr_name=repr_name)
 
     def constant(self, value, type=None):
         """
@@ -478,6 +706,7 @@ class ASTBuilder(object):
         """
         if type is None:
             type = self._infer_type(value)
+
         return ConstantNode(self.pos, type, value)
 
     def variable(self, type, name):
@@ -487,6 +716,19 @@ class ASTBuilder(object):
         """
         return Variable(self.pos, type, name)
 
+    def resolved_variable(self, array_type, name, element):
+        """
+        Creates a node that keeps the array operand information such as the
+        original array type, but references an actual element in the array.
+
+        :param type: original array type
+        :param name: original array's name
+        :param element: arbitrary expression that resolves some element in the
+                        array
+        """
+        return ResolvedVariable(self.pos, element.type, name,
+                                element=element, array_type=array_type)
+
     def cast(self, node, dest_type):
         "Cast node to the given destination type"
         return CastNode(self.pos, dest_type, node)
@@ -513,7 +755,7 @@ class ASTBuilder(object):
 
     def stridesvar(self, variable):
         "Return the strides variable for the given array operand"
-        return StridePointer(self.pos, minitypes.Py_ssize_t.pointer(), variable)
+        return StridePointer(self.pos, self.context.strides_type, variable)
 
     def stride(self, variable, index):
         "Return the stride of array operand `variable` at integer `index`"
@@ -575,6 +817,133 @@ class ASTBuilder(object):
         return NodeWrapper(self.context.getpos(opaque_node), type,
                            opaque_node, specialize_node_callback, **kwds)
 
+    #
+    ### Vectorization Functionality
+    #
+
+    def _vector_type(self, base_type, size):
+        return minitypes.VectorType(element_type=base_type, vector_size=size)
+
+    def vector_variable(self, variable, size):
+        "Return a vector variable for a data pointer variable"
+        type = self._vector_type(variable.type.dtype, size)
+
+        if size == 4:
+            name = 'xmm_%s' % variable.name
+        else:
+            name = 'ymm_%s' % variable.name
+
+        return VectorVariable(self.pos, type, name, variable=variable)
+
+    def vector_load(self, data_pointer, size):
+        "Load a SIMD vector of size `size` given an array operand variable"
+        type = self._vector_type(data_pointer.type.base_type, size)
+        return VectorLoadNode(self.pos, type, data_pointer, size=size)
+
+    def vector_store(self, data_pointer, vector_expr):
+        "Store a SIMD vector of size `size`"
+        assert data_pointer.type.base_type == vector_expr.type.element_type
+        return VectorStoreNode(self.pos, None, "=", data_pointer, vector_expr)
+
+    def vector_binop(self, operator, lhs, rhs):
+        "Perform a binary SIMD operation between two operands of the same type"
+        assert lhs.type == rhs.type, (lhs.type, rhs.type)
+        type = lhs.type
+        return VectorBinopNode(self.pos, type, operator, lhs=lhs, rhs=rhs)
+
+    def vector_unop(self, type, operator, operand):
+        return VectorUnopNode(self.pos, type, operator, operand)
+
+    def vector_const(self, type, constant):
+        return ConstantVectorNode(self.pos, type, constant=constant)
+
+    def noop_expr(self):
+        return NoopExpr(self.pos, type=None)
+
+class DynamicArgumentASTBuilder(ASTBuilder):
+    """
+    Create a function with a dynamic number of arguments. This means the
+    signature looks like
+
+        func(int *shape, float *data[n_ops], int *strides[n_ops])
+
+    To create minivect kernels supporting this signature, set the
+    astbuilder_cls attribute of Context to this class.
+    """
+
+    def data_pointer(self, variable):
+        if not hasattr(variable, 'data_pointer'):
+            temp =  self.temp(variable.type.dtype.pointer(),
+                              variable.name + "_data_temp")
+            variable.data_pointer = temp
+
+        return variable.data_pointer
+
+    def _create_data_pointer(self, function, argument, i):
+        variable = argument.variable
+
+        temp = self.data_pointer(variable)
+        p = self.index(function.data_pointers, self.constant(i))
+        p = self.cast(p, variable.type.dtype.pointer())
+        assmt = self.assign(temp, p)
+
+        function.body.stats.insert(0, assmt)
+        return temp
+
+    def stridesvar(self, variable):
+        "Return the strides variable for the given array operand"
+        if not hasattr(variable, 'strides_pointer'):
+            temp = self.temp(self.context.strides_type,
+                             variable.name + "_stride_temp")
+            variable.strides_pointer = temp
+
+        return variable.strides_pointer
+
+    def _create_strides_pointer(self, function, argument, i):
+        variable = argument.variable
+        temp = self.stridesvar(variable)
+        strides = self.index(function.strides_pointers, self.constant(i))
+        function.body.stats.insert(0, self.assign(temp, strides))
+        return temp
+
+    def function(self, name, body, args, shapevar=None, posinfo=None,
+                 omp_size=None):
+        function = super(DynamicArgumentASTBuilder, self).function(
+                        name, body, args, shapevar, posinfo, omp_size)
+
+        function.data_pointers = self.variable(
+                    minitypes.void.pointer().pointer(), 'data_pointers')
+        function.strides_pointers = self.variable(
+                    function.shape.type.pointer(), 'strides_pointer')
+
+        i = len(function.arrays) - 1
+        for argument in function.arrays[::-1]:
+            data_p = self._create_data_pointer(function, argument, i)
+            strides_p = self._create_strides_pointer(function, argument, i)
+
+            argument.data_pointer = data_p
+            argument.strides_pointer = strides_p
+
+            argument.used = False
+            i -= 1
+
+        argpos = 1
+        if posinfo:
+            argpos = 4
+
+        function.arguments.insert(argpos,
+                                  self.funcarg(function.strides_pointers))
+        function.arguments.insert(argpos,
+                                  self.funcarg(function.data_pointers))
+
+        self.create_function_type(function)
+        # print function.type
+        # print self.context.debug_c(
+        #        function, specializers.StridedSpecializer, type(self))
+        return function
+
+Context.astbuilder_cls = UndocClassAttribute(ASTBuilder)
+
 class Position(object):
     "Each node has a position which is an instance of this type."
 
@@ -594,7 +963,6 @@ class Node(miniutils.ComparableObjectMixin):
     is_expression = False
 
     is_statlist = False
-    is_scalar = False
     is_constant = False
     is_assignment = False
     is_unop = False
@@ -606,7 +974,10 @@ class Node(miniutils.ComparableObjectMixin):
     is_label = False
     is_temp = False
     is_statement = False
+    is_sizeof = False
+    is_variable = False
 
+    is_function = False
     is_funcarg = False
     is_array_funcarg = False
 
@@ -631,6 +1002,11 @@ class Node(miniutils.ComparableObjectMixin):
         visitor = minivisitor.PrintTree(context)
         visitor.visit(self)
 
+    @property
+    def children(self):
+        return [getattr(self, attr) for attr in self.child_attrs
+                    if getattr(self, attr) is not None]
+
     @property
     def comparison_objects(self):
         type = getattr(self, 'type', None)
@@ -647,7 +1023,7 @@ class Node(miniutils.ComparableObjectMixin):
     def __hash__(self):
         h = hash(type(self))
         for obj in self.comparison_objects:
-            h = h ^ hash(subtype)
+            h = h ^ hash(obj)
 
         return h
 
@@ -656,6 +1032,9 @@ class ExprNode(Node):
 
     is_expression = True
 
+    hoistable = False
+    need_temp = False
+
     def __init__(self, pos, type, **kwds):
         super(ExprNode, self).__init__(pos, **kwds)
         self.type = type
@@ -666,45 +1045,60 @@ class FunctionNode(Node):
     exceptions and success respectively.
 
     .. attribute:: shape
-            the broadcast shape for all operands
+
+        the broadcast shape for all operands
 
     .. attribute:: ndim
-            the ndim of the total broadcast' shape
+
+        the ndim of the total broadcast' shape
 
     .. attribute:: arguments
-            all array arguments
+
+        all array arguments
 
     .. attribute:: scalar arguments
+
         all non-array arguments
 
     .. attribute:: posinfo
+
         the position variables we can write to in case of an exception
+
+    .. attribute:: omp_size
+
+        the threshold of minimum data size needed before starting a parallel
+        section. May be overridden at any time before specialization time.
     """
 
+    is_function = True
+
     child_attrs = ['body', 'arguments', 'scalar_arguments']
 
     def __init__(self, pos, name, body, arguments, scalar_arguments,
-                 shape, posinfo, error_value, success_value):
+                 shape, posinfo, error_value, success_value, omp_size):
         super(FunctionNode, self).__init__(pos)
+        self.type = None # see ASTBuilder.create_function_type
         self.name = name
         self.body = body
+        self.arrays = [arg for arg in arguments if arg.type and arg.type.is_array]
         self.arguments = arguments
         self.scalar_arguments = scalar_arguments
         self.shape = shape
         self.posinfo = posinfo
         self.error_value = error_value
         self.success_value = success_value
+        self.omp_size = omp_size
 
         self.args = dict((v.name, v) for v in arguments)
         self.ndim = max(arg.type.ndim for arg in arguments
                                           if arg.type and arg.type.is_array)
 
-
 class FuncCallNode(ExprNode):
     """
     Call a function given a pointer or its name (FuncNameNode)
     """
 
+    inline = False
     child_attrs = ['func_or_pointer', 'args']
 
 class FuncNameNode(ExprNode):
@@ -754,6 +1148,8 @@ class FunctionArgument(ExprNode):
     child_attrs = ['variables']
     if_funcarg = True
 
+    used = True
+
     def __init__(self, pos, variable, variables):
         super(FunctionArgument, self).__init__(pos, variable.type)
         self.variables = variables
@@ -767,6 +1163,14 @@ class ArrayFunctionArgument(ExprNode):
     child_attrs = ['data_pointer', 'strides_pointer']
     is_array_funcarg = True
 
+    used = True
+
+    def __init__(self, pos, type, data_pointer, strides_pointer, **kwargs):
+        super(ArrayFunctionArgument, self).__init__(pos, type, **kwargs)
+        self.data_pointer = data_pointer
+        self.strides_pointer = strides_pointer
+        self.variables = [data_pointer, strides_pointer]
+
 class PrintNode(Node):
     "Print node for some arguments"
 
@@ -790,22 +1194,32 @@ class ForNode(Node):
 
     child_attrs = ['init', 'condition', 'step', 'body']
 
-    def __init__(self, pos, init, condition, step, body, is_tiled=False):
+    is_controlling_loop = False
+    is_tiling_loop = False
+
+    should_vectorize = False
+    is_fixup = False
+
+    def __init__(self, pos, init, condition, step, body, index=None):
         super(ForNode, self).__init__(pos)
         self.init = init
         self.condition = condition
         self.step = step
         self.body = body
-        self.is_tiled = False
+
+        self.index = index or init.lhs
 
 class IfNode(Node):
-    "An 'if' statement, see A for loop, see :py:class:`ASTBuilder.if_"
+    "An 'if' statement, see A for loop, see :py:class:`ASTBuilder.if_`"
 
-    child_attrs = ['cond', 'body']
+    child_attrs = ['cond', 'body', 'else_body']
+
+    should_vectorize = False
+    is_fixup = False
 
 class StatListNode(Node):
     """
-    A node to wrap multiple statements, see :py:class:`ASTBuilder.stats
+    A node to wrap multiple statements, see :py:class:`ASTBuilder.stats`
     """
     child_attrs = ['stats']
     is_statlist = True
@@ -830,7 +1244,6 @@ class NodeWrapper(ExprNode):
 
     is_node_wrapper = True
     is_constant_scalar = False
-    is_scalar = False
 
     child_attrs = []
 
@@ -860,8 +1273,8 @@ class NodeWrapper(ExprNode):
 class BinaryOperationNode(ExprNode):
     "Base class for binary operations"
     child_attrs = ['lhs', 'rhs']
-    def __init__(self, pos, type, lhs, rhs):
-        super(BinaryOperationNode, self).__init__(pos, type)
+    def __init__(self, pos, type, lhs, rhs, **kwds):
+        super(BinaryOperationNode, self).__init__(pos, type, **kwds)
         self.lhs, self.rhs = lhs, rhs
 
 class BinopNode(BinaryOperationNode):
@@ -869,8 +1282,8 @@ class BinopNode(BinaryOperationNode):
 
     is_binop = True
 
-    def __init__(self, pos, type, operator, lhs, rhs):
-        super(BinopNode, self).__init__(pos, type, lhs, rhs)
+    def __init__(self, pos, type, operator, lhs, rhs, **kwargs):
+        super(BinopNode, self).__init__(pos, type, lhs, rhs, **kwargs)
         self.operator = operator
 
     @property
@@ -880,8 +1293,8 @@ class BinopNode(BinaryOperationNode):
 class SingleOperandNode(ExprNode):
     "Base class for operations with one operand"
     child_attrs = ['operand']
-    def __init__(self, pos, type, operand):
-        super(SingleOperandNode, self).__init__(pos, type)
+    def __init__(self, pos, type, operand, **kwargs):
+        super(SingleOperandNode, self).__init__(pos, type, **kwargs)
         self.operand = operand
 
 class AssignmentExpr(BinaryOperationNode):
@@ -890,12 +1303,15 @@ class AssignmentExpr(BinaryOperationNode):
 class IfElseExprNode(ExprNode):
     child_attrs = ['cond', 'lhs', 'rhs']
 
+class PromotionNode(SingleOperandNode):
+    pass
+
 class UnopNode(SingleOperandNode):
 
     is_unop = True
 
-    def __init__(self, pos, type, operator, operand):
-        super(UnopNode, self).__init__(pos, type, operand)
+    def __init__(self, pos, type, operator, operand, **kwargs):
+        super(UnopNode, self).__init__(pos, type, operand, **kwargs)
         self.operator = operator
 
     @property
@@ -928,16 +1344,26 @@ class Variable(ExprNode):
     is_variable = True
     mangled_name = None
 
+    hoisted = False
+
     def __init__(self, pos, type, name, **kwargs):
         super(Variable, self).__init__(pos, type, **kwargs)
         self.name = name
+        self.array_type = None
 
     def __eq__(self, other):
-        return self.name == other.name
+        return isinstance(other, Variable) and self.name == other.name
 
     def __hash__(self):
         return hash(self.name)
 
+class ResolvedVariable(Variable):
+    child_attrs = ['element']
+
+    def __eq__(self, other):
+        return (isinstance(other, ResolvedVariable) and
+                self.element == other.element)
+
 class ArrayAttribute(Variable):
     "Denotes an attribute of array operands, e.g. the data or stride pointers"
     def __init__(self, pos, type, arrayvar):
@@ -960,8 +1386,6 @@ class StridePointer(ArrayAttribute):
 class TempNode(Variable):
     "A temporary of a certain type"
 
-    child_attrs = ['rhs']
-
     is_temp = True
 
     def __eq__(self, other):
@@ -974,7 +1398,13 @@ class OpenMPLoopNode(Node):
     """
     Execute a loop in parallel.
     """
-    child_attrs = ['for_node', 'if_clause']
+    child_attrs = ['for_node', 'if_clause', 'lastprivates', 'privates']
+
+class OpenMPConditionalNode(Node):
+    """
+    Execute if_body if _OPENMP, otherwise execute else_body.
+    """
+    child_attrs = ['if_body', 'else_body']
 
 class PragmaForLoopNode(Node):
     """
@@ -1040,3 +1470,28 @@ class LabelNode(ExprNode):
         super(LabelNode, self).__init__(pos, None)
         self.name = name
         self.mangled_name = None
+
+class NoopExpr(ExprNode):
+    "Do nothing expression"
+
+#
+### Vectorization Functionality
+#
+
+class VectorVariable(Variable):
+    child_attrs = ['variable']
+
+class VectorLoadNode(SingleOperandNode):
+    "Load a SIMD vector"
+
+class VectorStoreNode(BinopNode):
+    "Store a SIMD vector"
+
+class VectorBinopNode(BinopNode):
+    "Binary operation on SIMD vectors"
+
+class VectorUnopNode(SingleOperandNode):
+    "Unary operation on SIMD vectors"
+
+class ConstantVectorNode(ExprNode):
+    "Load the constant into the vector register"
\ No newline at end of file
