@@ -27,21 +27,40 @@ Traceback (most recent call last):
 InvalidTypeSpecification: Step may only be provided once, and only in the first or last dimension.
 """
 
-__all__ = ['Py_ssize_t', 'void', 'char', 'uchar', 'int_', 'long_', 'bool_', 'object_',
+__all__ = ['Py_ssize_t', 'void', 'char', 'uchar', 'short', 'ushort',
+           'int_', 'uint', 'long_', 'ulong', 'longlong', 'ulonglong',
+           'size_t', 'npy_intp', 'c_string_type', 'bool_', 'object_',
            'float_', 'double', 'longdouble', 'float32', 'float64', 'float128',
            'int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64',
            'complex64', 'complex128', 'complex256']
 
+import sys
+import math
+import copy
+import ctypes
+import textwrap
+
 try:
-    import llvm.core as lc
+    import llvm.core
+    from llvm import core as lc
 except ImportError:
-    lc = None
+    llvm = None
 
 import miniutils
 import minierror
 
+# Check below taken from Numba
+if sys.maxint > 2**33:
+    _plat_bits = 64
+else:
+    _plat_bits = 32
+
+
 class TypeMapper(object):
     """
+    Maps foreign types to minitypes. Users of minivect should implement
+    this and pass it to :py:class:`minivect.miniast.Context`.
+
     >>> import miniast
     >>> context = miniast.Context()
     >>> miniast.typemapper = TypeMapper(context)
@@ -92,10 +111,12 @@ class TypeMapper(object):
     >>> tm.promote_types(int_[:, :], object_[:, ::1])
     PyObject *[:, :]
     """
+
     def __init__(self, context):
         self.context = context
 
     def map_type(self, opaque_type):
+        "Map a foreign type to a minitype"
         if opaque_type.is_int:
             return int_
         elif opaque_type.is_float:
@@ -117,12 +138,21 @@ class TypeMapper(object):
 
     def from_python(self, value):
         "Get a type from a python value"
+        np = sys.modules.get('numpy', None)
+
         if isinstance(value, float):
             return double
         elif isinstance(value, (int, long)):
             return int_
         elif isinstance(value, complex):
             return complex128
+        elif isinstance(value, str):
+            return c_string_type
+        elif np and isinstance(value, np.ndarray):
+            dtype = map_dtype(value.dtype)
+            return ArrayType(dtype, value.ndim,
+                             is_c_contig=value.flags['C_CONTIGUOUS'],
+                             is_f_contig=value.flags['F_CONTIGUOUS'])
         else:
             return object_
             # raise minierror.UnmappableTypeError(type(value))
@@ -151,11 +181,106 @@ class TypeMapper(object):
             return object_
         elif type1.is_numeric and type2.is_numeric:
             return self.promote_numeric(type1, type2)
-        elif type1.is_array and type2:
+        elif type1.is_array and type2.is_array:
             return self.promote_arrays(type1, type2)
         else:
             raise minierror.UnpromotableTypeError((type1, type2))
 
+def map_dtype(dtype):
+    """
+    Map a NumPy dtype to a minitype.
+
+    >>> import numpy as np
+    >>> map_dtype(np.dtype(np.int32))
+    int32
+    >>> map_dtype(np.dtype(np.int64))
+    int64
+    >>> map_dtype(np.dtype(np.object))
+    PyObject *
+    >>> map_dtype(np.dtype(np.float64))
+    double
+    >>> map_dtype(np.dtype(np.complex128))
+    complex128
+    """
+    item_idx = int(math.log(dtype.itemsize, 2))
+    if dtype.kind == 'i':
+        return [int8, int16, int32, int64][item_idx]
+    elif dtype.kind == 'u':
+        return [uint8, uint16, uint32, uint64][item_idx]
+    elif dtype.kind == 'f':
+        if dtype.itemsize == 2:
+            pass # half floats not supported yet
+        elif dtype.itemsize == 4:
+            return float32
+        elif dtype.itemsize == 8:
+            return float64
+        elif dtype.itemsize == 16:
+            return float128
+    elif dtype.kind == 'b':
+        return int8
+    elif dtype.kind == 'c':
+        if dtype.itemsize == 8:
+            return complex64
+        elif dtype.itemsize == 16:
+            return complex128
+        elif dtype.itemsize == 32:
+            return complex256
+    elif dtype.kind == 'O':
+        return object_
+
+def create_dtypes():
+    import numpy as np
+
+    minitype2dtype = {
+        int8     : np.int8,
+        int16    : np.int16,
+        int32    : np.int32,
+        int64    : np.int64,
+        uint8    : np.uint8,
+        uint16   : np.uint16,
+        uint32   : np.uint32,
+        uint64   : np.uint64,
+
+        float_   : np.float32,
+        double   : np.float64,
+        longdouble: np.longdouble,
+
+        short    : np.dtype('h'),
+        int_     : np.dtype('i'),
+        long_    : np.dtype('l'),
+        longlong : np.longlong,
+        ushort   : np.dtype('H'),
+        uint     : np.dtype('I'),
+        ulong    : np.dtype('L'),
+        ulonglong: np.ulonglong,
+
+        complex64: np.complex64,
+        complex128: np.complex128,
+        complex256: getattr(np, 'complex256', None),
+
+        object_  : np.object,
+    }
+
+    return dict((k, np.dtype(v)) for k, v in minitype2dtype.iteritems())
+
+_dtypes = None
+def map_minitype_to_dtype(type):
+    global _dtypes
+
+    if _dtypes is None:
+        _dtypes = create_dtypes()
+
+    if type.is_array:
+        type = type.dtype
+
+    dtype = _dtypes[type]
+    assert dtype is not None, "dtype not supported in this numpy build"
+    return dtype
+
+NONE_KIND = 0
+INT_KIND = 1
+FLOAT_KIND = 2
+COMPLEX_KIND = 3
 
 class Type(miniutils.ComparableObjectMixin):
     """
@@ -183,6 +308,8 @@ class Type(miniutils.ComparableObjectMixin):
     is_complex = False
     is_void = False
 
+    kind = NONE_KIND
+
     subtypes = []
 
     def __init__(self, **kwds):
@@ -221,6 +348,9 @@ class Type(miniutils.ComparableObjectMixin):
         return (type(self) is type(other) and
                 self.comparison_type_list == other.comparison_type_list)
 
+    def __ne__(self, other):
+        return not self == other
+
     def __hash__(self):
         h = hash(type(self))
         for subtype in self.comparison_type_list:
@@ -229,6 +359,11 @@ class Type(miniutils.ComparableObjectMixin):
         return h
 
     def __getitem__(self, item):
+        """
+        Support array type creation by slicing, e.g. double[:, :] specifies
+        a 2D strided array of doubles. The syntax is the same as for
+        Cython memoryviews.
+        """
         assert isinstance(item, (tuple, slice))
 
         def verify_slice(s):
@@ -256,6 +391,9 @@ class Type(miniutils.ComparableObjectMixin):
             verify_slice(item)
             return ArrayType(self, 1, is_c_contig=bool(item.step))
 
+    def declare(self):
+        return str(self)
+
     def to_llvm(self, context):
         "Get a corresponding llvm type from this type"
         return context.to_llvm(self)
@@ -266,18 +404,32 @@ class Type(miniutils.ComparableObjectMixin):
         return getattr(type(self), attr)
 
 class ArrayType(Type):
+    """
+    An array type. ArrayType may be sliced to obtain a subtype:
+
+    >>> double[:, :, ::1][1:]
+    double[:, ::1]
+    >>> double[:, :, ::1][:-1]
+    double[:, :]
+
+    >>> double[::1, :, :][:-1]
+    double[::1, :]
+    >>> double[::1, :, :][1:]
+    double[:, :]
+    """
 
     is_array = True
     subtypes = ['dtype']
 
     def __init__(self, dtype, ndim, is_c_contig=False, is_f_contig=False,
-                 inner_contig=False):
+                 inner_contig=False, broadcasting=None):
         super(ArrayType, self).__init__()
         self.dtype = dtype
         self.ndim = ndim
         self.is_c_contig = is_c_contig
         self.is_f_contig = is_f_contig
         self.inner_contig = inner_contig or is_c_contig or is_f_contig
+        self.broadcasting = broadcasting
 
     @property
     def comparison_type_list(self):
@@ -287,8 +439,9 @@ class ArrayType(Type):
         raise Exception("You probably want a pointer type to the dtype")
 
     def to_llvm(self, context):
-        raise Exception("Obtain a pointer to the dtype and convert that "
-                        "to an LLVM type")
+        # raise Exception("Obtain a pointer to the dtype and convert that "
+        #                 "to an LLVM type")
+        return context.to_llvm(self)
 
     def __repr__(self):
         axes = [":"] * self.ndim
@@ -299,6 +452,45 @@ class ArrayType(Type):
 
         return "%s[%s]" % (self.dtype, ", ".join(axes))
 
+    @property
+    def strided(self):
+        type = copy.copy(self)
+        type.is_c_contig = False
+        type.is_f_contig = False
+        type.inner_contig = False
+        type.broadcasting = None
+        return type
+
+    def __getitem__(self, index):
+        assert isinstance(index, slice)
+        assert index.step is None
+        assert index.start is not None or index.stop is not None
+
+        start = 0
+        stop = self.ndim
+        if index.start is not None:
+            start = index.start
+        if index.stop is not None:
+            stop = index.stop
+
+        ndim = len(range(self.ndim)[start:stop])
+
+        if ndim == 0:
+            type = self.dtype
+        elif ndim > 0:
+            type = self.strided
+            type.ndim = ndim
+            type.is_c_contig = self.is_c_contig and stop == self.ndim
+            type.is_f_contig = self.is_f_contig and start == 0
+            type.inner_contig = type.is_c_contig or type.is_f_contig
+            if type.broadcasting:
+                type.broadcasting = self.broadcasting[start:stop]
+        else:
+            raise IndexError(index, ndim)
+
+        return type
+
+
 class PointerType(Type):
     is_pointer = True
     subtypes = ['base_type']
@@ -311,7 +503,12 @@ class PointerType(Type):
         return "%s *%s" % (self.base_type, " ".join(self.qualifiers))
 
     def to_llvm(self, context):
-        return lc.Type.pointer(self.base_type.to_llvm(context))
+        if self.base_type.is_void:
+            llvm_base_type = int_.to_llvm(context)
+        else:
+            llvm_base_type = self.base_type.to_llvm(context)
+
+        return llvm.core.Type.pointer(llvm_base_type)
 
 class CArrayType(Type):
     is_carray = True
@@ -326,7 +523,7 @@ class CArrayType(Type):
         return "%s[%d]" % (self.base_type, self.length)
 
     def to_llvm(self, context):
-        return lc.Type.array(self.base_type.to_llvm(context), self.size)
+        return llvm.core.Type.array(self.base_type.to_llvm(context), self.size)
 
 class TypeWrapper(Type):
     is_typewrapper = True
@@ -345,13 +542,14 @@ class TypeWrapper(Type):
 
 class NamedType(Type):
     name = None
+
     def __eq__(self, other):
         return isinstance(other, NamedType) and self.name == other.name
 
     def __repr__(self):
         if self.qualifiers:
             return "%s %s" % (self.name, " ".join(self.qualifiers))
-        return self.name
+        return str(self.name)
 
 class BoolType(NamedType):
     is_bool = True
@@ -389,6 +587,8 @@ class IntType(NumericType):
     rank = 4
     itemsize = 4
 
+    kind = INT_KIND
+
     def to_llvm(self, context):
         if self.itemsize == 1:
             return lc.Type.int(8)
@@ -400,9 +600,29 @@ class IntType(NumericType):
             assert self.itemsize == 8, self
             return lc.Type.int(64)
 
+    def declare(self):
+        if self.name.endswith(('16', '32', '64')):
+            return self.name + "_t"
+        else:
+            return str(self)
+
 class FloatType(NumericType):
     is_float = True
 
+    kind = FLOAT_KIND
+
+    def declare(self):
+        if self.itemsize == 4:
+            return "float"
+        elif self.itemsize == 8:
+            return "double"
+        else:
+            return str(self)
+
+    @property
+    def comparison_type_list(self):
+        return self.subtype_list + [self.itemsize]
+
     def to_llvm(self, context):
         if self.itemsize == 4:
             return lc.Type.float()
@@ -417,12 +637,29 @@ class ComplexType(NumericType):
     is_complex = True
     subtypes = ['base_type']
 
+    kind = COMPLEX_KIND
+
 class Py_ssize_t_Type(IntType):
     is_py_ssize_t = True
     name = "Py_ssize_t"
     rank = 9
     signed = True
 
+    def __init__(self, **kwds):
+        super(Py_ssize_t_Type, self).__init__(**kwds)
+        self.itemsize = _plat_bits / 8
+
+class NPyIntp(IntType):
+    is_numpy_intp = True
+    name = "npy_intp"
+    rank = 10
+
+    def __init__(self, **kwds):
+        super(NPyIntp, self).__init__(**kwds)
+        import numpy as np
+        ctypes_array = np.empty(0).ctypes.strides
+        self.itemsize = ctypes.sizeof(ctypes_array._type_)
+
 class CharType(IntType):
     is_char = True
     name = "char"
@@ -460,11 +697,52 @@ class FunctionType(Type):
     is_vararg = False
 
     def to_llvm(self, context):
+        assert self.return_type is not None
         return lc.Type.function(self.return_type.to_llvm(context),
                                 [arg_type.to_llvm(context)
                                     for arg_type in self.args],
                                 self.is_vararg)
 
+    def __str__(self):
+        args = map(str, self.args)
+        if self.is_vararg:
+            args.append("...")
+
+        return "%s (*)(%s)" % (self.return_type, ", ".join(args))
+
+class VectorType(Type):
+    subtypes = ['element_type']
+    is_vector = True
+    vector_size = None
+
+    def __init__(self, element_type, vector_size, **kwds):
+        super(VectorType, self).__init__(**kwds)
+        assert ((element_type.is_int or element_type.is_float) and
+                element_type.itemsize in (4, 8)), element_type
+        self.element_type = element_type
+        self.vector_size = vector_size
+
+    def to_llvm(self, context):
+        return lc.Type.vector(self.element_type.to_llvm(context),
+                              self.vector_size)
+
+    @property
+    def comparison_type_list(self):
+        return self.subtype_list + [self.vector_size]
+
+    def __str__(self):
+        itemsize = self.element_type.itemsize
+        if self.element_type.is_float:
+            if itemsize == 4:
+                return '__m128'
+            else:
+                return '__m128d'
+        else:
+            if itemsize == 4:
+                return '__m128i'
+            else:
+                raise NotImplementedError
+
 #
 ### Internal types
 #
@@ -475,6 +753,7 @@ void = VoidType()
 ### Public types
 #
 Py_ssize_t = Py_ssize_t_Type()
+npy_intp = NPyIntp()
 size_t = IntType(name="size_t", rank=8.5, itemsize=8, signed=False)
 char = CharType(name="char")
 short = IntType(name="short", rank=2, itemsize=2)
@@ -513,6 +792,40 @@ complex128 = ComplexType(name="complex128", base_type=float64,
 complex256 = ComplexType(name="complex256", base_type=float128,
                          rank=20, itemsize=32)
 
+def get_utility():
+    import numpy
+
+    return textwrap.dedent("""\
+    #include <stdint.h>
+
+    #ifndef HAVE_LONGDOUBLE
+        #define HAVE_LONGDOUBLE %d
+    #endif
+
+    typedef struct {
+        float real;
+        float imag;
+    } complex64;
+
+    typedef struct {
+        double real;
+        double imag;
+    } complex128;
+
+    #if HAVE_LONGDOUBLE
+    typedef struct {
+        long double real;
+        long double imag;
+    } complex256;
+    #endif
+
+    typedef float float32;
+    typedef double float64;
+    #if HAVE_LONGDOUBLE
+    typedef long double float128;
+    #endif
+    """ % hasattr(numpy, 'complex256'))
+
 if __name__ == '__main__':
     import doctest
     doctest.testmod()
