@@ -1,9 +1,13 @@
 import opcode
 import ctypes
+
 from numpy import complex64, complex128
 
+from .minivect import miniast, minitypes
+from . import _numba_types
+
 def itercode(code):
-    """Return a generator of byte-offset, opcode, and argument 
+    """Return a generator of byte-offset, opcode, and argument
     from a byte-code-string
     """
     i = 0
@@ -35,6 +39,26 @@ def debugout(*args):
         print("debugout (non-translated): %s" % (''.join((str(arg)
                                                           for arg in args)),))
 
+
+class NumbaContext(miniast.LLVMContext):
+    # debug = True
+    # debug_elements = True
+
+    # Accept dynamic arguments
+    astbuilder_cls = miniast.DynamicArgumentASTBuilder
+
+    shape_type = minitypes.npy_intp.pointer()
+    strides_type = shape_type
+
+    def init(self):
+        self.astbuilder = self.astbuilder_cls(self)
+        self.typemapper = _numba_types.NumbaTypeMapper(self)
+
+def get_minivect_context():
+    return NumbaContext()
+
+context = get_minivect_context()
+
 # NOTE: The following ctypes structures were inspired by Joseph
 # Heller's response to python-list question about ctypes complex
 # support.  In that response, he said these were only suitable for
@@ -80,4 +104,4 @@ class Complex64 (ctypes.Structure, ComplexMixin):
 
 class Complex128 (ctypes.Structure, ComplexMixin):
     _fields_ = [('real', ctypes.c_double), ('imag', ctypes.c_double)]
-    _numpy_ty_ = complex128
+    _numpy_ty_ = complex128
\ No newline at end of file
